
#ifdef _KERNEL

#include <dev/sound/pcm/sound.h>
#include "feeder_if.h"
SND_DECLARE_FILE("$FreeBSD$");

#endif /* _KERNEL */

MALLOC_DEFINE(M_SINCFEEDER, "sincfeed", "pcm sinc feeder");

#define SINC_ASSERT(x, y...) 						      \
	do {								      \
    	if (!(x)) {							      \
	    printf y;							      \
	    abort();							      \
	    }								      \
	} while (0)

#ifdef SINC_DEBUG
#define SINC_TRACE(x...)  fprintf(stderr, x) 
#else
#define SINC_TRACE(x...)
#endif

#include "feeder_sinc.h"

struct feed_sinc_info;

typedef int (*sinc_filter_method)(struct feed_sinc_info*, uint32_t, int16_t*);

#define FEEDBUFSZ	8192
#define MAX_COEFF	((2 * SIGMA_MAX_WIDTH - 1) + (3)) /* + 3 == mmx padding */

#define GAIN		240
#define MAX_GAIN	256

#define MAX_Q		SIGMA_TABLE_SIZE

struct feed_sinc_info {
	int src;			/* src rate */
	int dst;			/* dst_rate */

	size_t	 n_coeff;		/* number of filter coefficients */
	int16_t  coeff[MAX_COEFF];	/* filter coefficients */

	int16_t  buffer[FEEDBUFSZ];	/* input buffer */
	uint16_t buffer_ticks;		/* number of available samples */
	uint16_t buffer_pos;		/* next available sample */
	uint8_t	 channels;		/* 1 = mono, 2 = stereo */

	uint8_t	 quality;		/* quality (0-MAX_Q) -> filter size */
	uint8_t  impl;			/* impl 0 = int, 1 = mmx */
};

#define BYTES_PER_SAMPLE	2
#define samples_per_tick(info)	(info->channels)
#define bytes_per_tick(info)	(info->channels * BYTES_PER_SAMPLE)
#define history_ticks(info)	(info->n_coeff)
#define history_samples(info)	(history_ticks(info) * samples_per_tick(info))
#define history_bytes(info)	(history_ticks(info) * bytes_per_tick(info))

#define max_buffer_bytes(info)	(FEEDBUFSZ - history_bytes(info))
#define max_buffer_ticks(info)	(max_buffer_bytes(info) / bytes_per_tick(info))

/* Clip operation - cheap in comparison to multiply adds */
#define clip16(x) 							      \
if (x > 32767) { 							      \
	fprintf(stderr, "Clipping %d (line %s:%d)\n", x, __FILE__, __LINE__); \
	x = 32767;  							      \
} else if (x < -32767) { 						      \
	fprintf(stderr, "Clipping %d (line %s:%d)\n", x, __FILE__, __LINE__); \
	x = -32767; 							      \
}

/*
#undef clip16
#define clip16(x)
*/

static void
normalize(int16_t *sb, uint32_t n)
{
	uint32_t i;
	int32_t num, denom, total, tmp;

	/* Input samples are in range -32768--32767. */
	for (i = 0, total = 0; i < n; i++) {
		total += sb[i];
	}

	num = 32767;
	denom = total;

	while (denom > 32767) {
		denom >>= 1;
		num >>= 1;
	}

	fprintf(stderr, "num = %d denom %d\n", num, denom);

	if (denom == 0)
		denom = 1;

	for (i = 0; i < n; i++) {
		tmp = sb[i] * num / denom;
		SINC_ASSERT(tmp <= 32767 && tmp >= -32768,
			    ("bad coefficient %d\n", tmp));
		sb[i] = tmp;
	}
}

static void
init_sinc_filter(struct feed_sinc_info* info, int sigma_index, int sinc_index)
{
	size_t mid_coeff, i;
	int32_t tmp;

	memset(info->coeff, 0, sizeof(info->coeff[0]) * MAX_COEFF);

	mid_coeff = sigma_tab[sigma_index].ncoeff - 1;
	info->n_coeff = 1 + 2 * mid_coeff;

	tmp = sigma_tab[sigma_index].coeff[0] *
		sinc_tab[sinc_index].coeff[0];
	tmp >>= sigma_precision;
	SINC_ASSERT(tmp <= 32767 && tmp >= -32768,
		    ("bad coefficient %d\n", tmp));
	info->coeff[mid_coeff] = tmp;

	/*
	 * The tabulated coefficients represent 1/2 of filter width,
	 * calculate convoluted coefficients, scale, and mirror.
	 */
	for (i = 1; i <= mid_coeff; i++) {
		tmp = sigma_tab[sigma_index].coeff[i] *
			sinc_tab[sinc_index].coeff[i];
		tmp >>= sigma_precision;
		SINC_ASSERT(tmp <= 32767 && tmp >= -32768,
			    ("bad coefficient %d\n", tmp));

		info->coeff[mid_coeff + i] = tmp;
		info->coeff[mid_coeff - i] = tmp;
	}
	normalize(info->coeff, info->n_coeff);
	for (i = 0; i < info->n_coeff; i++) {
		tmp = info->coeff[i] * GAIN / MAX_GAIN;
		SINC_ASSERT(tmp <= 32767 && tmp >= -32768,
			    ("bad coefficient %d\n", tmp));
		info->coeff[i] = tmp;
	}

	if (info->impl == 1) {
	    while (info->n_coeff & 0x03) {
		info->coeff[info->n_coeff] = 0;
		info->n_coeff++;
	    }
	}

	for (i = 0; i < info->n_coeff; i++) {
		fprintf(stderr, "a[%d] = %d\n", i, info->coeff[i]);
	}
}

static int
feed_sinc_setup(struct pcm_feeder *f)
{
	struct feed_sinc_info *info = f->data;
	size_t sigma_index, sinc_index, f_l, f_h;

	if (info->src > info->dst) {
		f_h = info->src;
		f_l = info->dst;
	} else {
		f_h = info->dst;
		f_l = info->src;
	}
	sinc_index = sinc_table_index(f_l, f_h);

#if 0
	if (sinc_index > 2)
		sinc_index -= 2;
	else
		sinc_index = 0;
#endif /* 0 */

	sigma_index = info->quality * SIGMA_TABLE_SIZE / MAX_Q - 1;
	if (sigma_index >= SIGMA_TABLE_SIZE)
		sigma_index = SIGMA_TABLE_SIZE - 1;


	fprintf(stderr,
		"Sinc index = %d cutoff %d quality %d sigma index %d\n",
		sinc_index, sinc_tab[sinc_index].f * f_h / sinc_f_scale,
		info->quality, sigma_index);

	init_sinc_filter(info, sigma_index, sinc_index);

	return 0;
}

static int
feed_sinc_init(struct pcm_feeder *f)
{
	struct feed_sinc_info *info;

	info = malloc(sizeof(*info), M_SINCFEEDER, M_NOWAIT | M_ZERO);
	if (info == NULL)
		return ENOMEM;

	info->src = 8000;
	info->dst = 8000;
	info->quality = 1;
	info->channels = 2;
	info->buffer_ticks = 0;
	info->buffer_pos = 0;

	f->data = info;

	return feed_sinc_setup(f);
}

static int
feed_sinc_free(struct pcm_feeder *f)
{
	struct feed_sinc_info *info = f->data;

	if (info) {
		free(info, M_SINCFEEDER);
		f->data = NULL;
	}
	return 0;
}

static int
feed_sinc_set(struct pcm_feeder *f, int what, int value)
{
	struct feed_sinc_info *info = f->data;

	switch(what) {
	case FEEDSINC_SRC:
		if (value < 0 || value > 48000)
			return -1;
		info->src = value;
		break;
	case FEEDSINC_DST:
		if (value < 0 || value > 48000)
			return -1;
		info->dst = value;
		break;
	case FEEDSINC_QUALITY:
		if (value < 0 || value > (int)MAX_Q)
			return -1;
		info->quality = value;
		break;
	case FEEDSINC_IMPL:
		if (value < 0 || value > 1)
			return -1;
		info->impl = value;
		break;
	default:
		return -1;
	}
	return feed_sinc_setup(f);
}

static int
feed_sinc_get(struct pcm_feeder *f, int what)
{
	struct feed_sinc_info *info = f->data;

	switch(what) {
	case FEEDSINC_SRC:
		return info->src;
	case FEEDSINC_DST:
		return info->dst;
	case FEEDSINC_QUALITY:
		return info->quality;
	case FEEDSINC_IMPL:
		return info->impl;
	default:
		return -1;
	}
	return -1;
}

static int
sinc_filter(struct feed_sinc_info *info, uint32_t ticks, int16_t* dst)
{
	uint32_t s, es, i, delta, delta1, delta2, delta3;
	int16_t *src;
	int32_t tmp, x, m[4];
	int32_t xb[16];

	src = info->buffer + info->buffer_pos * samples_per_tick(info);
	es = ticks * samples_per_tick(info);
	/* This loop is unrolled - we do 16 ops in a round, this seems to
	 * be about the optimum limit on the p4 box being used for testing.
	 * 32 decreases performance slightly.  It doesn't really matter that
	 * much.  This would be ripe for SIMD manipulation.
	 */
	s = 0;
	while (es - s > 16) {
	    memset(xb, 0, sizeof(xb));
	    for (i = 0; info->n_coeff - i >= 4; i += 4) {
		delta = i * info->channels + s;
		delta1 = (i + 1) * info->channels + s;
		delta2 = (i + 2) * info->channels + s;
		delta3 = (i + 3) * info->channels + s;		
		m[0] = info->coeff[i];
		m[1] = info->coeff[i + 1];
		m[2] = info->coeff[i + 2];
		m[3] = info->coeff[i + 3];
#define FOUR_OPS(n) 							      \
xb[ n] += (src[ n + delta] * m[0] >> 8) + (src[ n + delta1] * m[1] >> 8) +    \
	  (src[ n + delta2] * m[2] >> 8) + (src[ n + delta3] * m[3] >> 8);
		FOUR_OPS(0);
		FOUR_OPS(1);
		FOUR_OPS(2);
		FOUR_OPS(3);
		FOUR_OPS(4);
		FOUR_OPS(5);
		FOUR_OPS(6);
		FOUR_OPS(7);
		FOUR_OPS(8);
		FOUR_OPS(9);
		FOUR_OPS(10);
		FOUR_OPS(11);
		FOUR_OPS(12);
		FOUR_OPS(13);
		FOUR_OPS(14);
		FOUR_OPS(15);
	    }

	    for (; i < info->n_coeff; i++) {
		delta = i * info->channels + s;
		m[0] = info->coeff[i];
#define ONE_OPS(n)							      \
		xb[ n] += (src[ n + delta] * m[0] >> 8);
		ONE_OPS(0);
		ONE_OPS(1);
		ONE_OPS(2);
		ONE_OPS(3);
		ONE_OPS(4);
		ONE_OPS(5);
		ONE_OPS(6);
		ONE_OPS(7);
		ONE_OPS(8);
		ONE_OPS(9);
		ONE_OPS(10);
		ONE_OPS(11);
		ONE_OPS(12);
		ONE_OPS(13);
		ONE_OPS(14);
		ONE_OPS(15);
	    }
#define pack16(n) xb[n] >>= 7; clip16(xb[n]); dst[s + n] = xb[n];
	    pack16(0);
	    pack16(1);
	    pack16(2);
	    pack16(3);
	    pack16(4);
	    pack16(5);
	    pack16(6);
	    pack16(7);
	    pack16(8);
	    pack16(9);
	    pack16(10);
	    pack16(11);
	    pack16(12);
	    pack16(13);
	    pack16(14);
	    pack16(15);
	    s += 16;
	}

	for (;s < es; s++) {
	    tmp = 0;
	    for (i = 0; i < info->n_coeff; i++) {
		x = src[s + i * info->channels] * info->coeff[i];
		tmp += x;
	    }
	    tmp >>= 15;
	    clip16(tmp);
	    dst[s] = tmp;
	}

	info->buffer_pos += ticks;
	SINC_ASSERT(info->buffer_pos <= info->buffer_ticks, 
		    ("Buffer overrun"));
	return ticks;
}


int
sinc_filter_mmx(struct feed_sinc_info *info, uint32_t ticks, int16_t* dst);

#if 0
void
test_round(int16_t* filter_coeff, int16_t* stereo_samples, int n_samples, int32_t* eresults);

void
test_round(int16_t* filter_coeff, int16_t* stereo_samples, int n_samples, int32_t* eresults)
{
	int i, j, k;
	int32_t results[4];

	results[0] = results[1] = results[2] = results[3] = 0;
	for (k = 0; k < 2; k++) {
	    for (j = 0; j < n_samples ; j++) {
		for (i = 0; i < 2; i++) {
		    results[i + k * 2] += filter_coeff[j] * stereo_samples[i + j * 2 + k * 2] >> 8;
		}
	    }
	}
	for (j = 0; j < 4; j++) {
	    if (abs(results[j] - eresults[j]) > 8) {
		for (j = 0; j < 4; j++) {
		    fprintf(stderr, "	%5d. %8d %8d %c\n",
			    j, results[j], eresults[j], (results[j] == eresults[j]) ? ' ': '*');
		}
		return;
	    }
	}
}
#endif

int
sinc_filter_mmx(struct feed_sinc_info *info, uint32_t ticks, int16_t* dst)
{
	uint32_t s, es, i, ms;
	int16_t *src;
	int32_t tmp[4], x;
	int32_t xb[4];

	src = info->buffer + info->buffer_pos * samples_per_tick(info);
	es = ticks * samples_per_tick(info);
	s = 0;
	ms = es & ~3;
	while (s < ms) {
	    memset(xb, 0, sizeof(xb));
		for (i = 0; i < info->n_coeff; i += 4) {
			__asm__(
				"\n\t movq %2, %%mm0"		/* mm0 f3 f2 f1 f0 */
				"\n\t movq %3, %%mm2"		/* mm2 r1 l1 r0 l0 */
				"\n\t movq %%mm0, %%mm1"	/* mm1 f3 f2 f1 f0 */
				"\n\t movq %4, %%mm3"		/* mm3 r2 l2 r1 l1 */
				"\n\t movq %5, %%mm4"		/* mm4 r3 l3 r2 l2 */
				"\n\t movq %6, %%mm5"		/* mm5 r4 l4 r3 l3 */

				"\n\t movq %%mm2, %%mm6"	/* mm6 r1 l1 r0 l0 */
				"\n\t movq %%mm4, %%mm7"	/* mm7 r3 l3 r2 l2 */
				"\n\t punpckldq %%mm0, %%mm0"	/* mm0 f1 f0 f1 f0 */
				"\n\t punpckhdq %%mm1, %%mm1"	/* mm1 f3 f2 f3 f2 */

				"\n\t punpcklwd %%mm3, %%mm2"	/* mm2 r1 r0 l1 l0 */
				"\n\t punpcklwd %%mm5, %%mm4"	/* mm4 r3 r2 l3 l2 */
				"\n\t punpckhwd %%mm3, %%mm6"	/* mm6 r2 r1 l2 l1 */
				"\n\t punpckhwd %%mm5, %%mm7"	/* mm7 r4 r3 l4 l3 */
				"\n\t pmaddwd   %%mm0, %%mm2"	/* mm2 (f1.r1 + f0.r0) (f1.l1 + f0.l0) */
				"\n\t pmaddwd   %%mm1, %%mm4"	/* mm4 (f3.r3 + f2.r2) (f3.l3 + f2.l2) */
				"\n\t psrad	$8, %%mm2"	/* mm2 >>= 8 */
				"\n\t psrad	$8, %%mm4"	/* mm4 >>= 8 */
				"\n\t pmaddwd   %%mm0, %%mm6"	/* mm6 (f1.r2 + f0.r1) (f1.l2 + f0.l1) */
				"\n\t pmaddwd   %%mm1, %%mm7"	/* mm7 (f3.r4 + f2.r3) (f3.l4 + f2.l3) */
				"\n\t psrad	$8, %%mm6"	/* mm6 >>= 8 */
				"\n\t psrad	$8, %%mm7"	/* mm7 >>= 8 */
				"\n\t paddd     %%mm2, %%mm4"	/* mm4 (f3.r3 + f2.r2 + f1.r1 + f0.r0 ) (f3.l3 + f2.l2 + f1.l1 + f0.l0) */
				"\n\t paddd     %%mm6, %%mm7"	/* mm7 (f3.r4 + f2.r3 + f1.r2 + f0.r1 ) (f3.l4 + f2.l3 + f1.l2 + f0.l1) */
				"\n\t movq	%0, %%mm0"
				"\n\t movq	%1, %%mm1"
				"\n\t paddd	%%mm0, %%mm4"
				"\n\t paddd	%%mm1, %%mm7"
				"\n\t movq	%%mm4, %0"
				"\n\t movq	%%mm7, %1"
				: "=m" (xb[0]), "=m" (xb[2])
				: "m" (info->coeff[i]), "m" (src[s + 2 * i]), "m" (src[s + 2 + 2 * i]), "m" (src[s + 4 + 2 * i]), "m" (src[s + 6 + 2 *i])
				);
		}

		__asm__(
		    "\n\t movq %1, %%mm0"
		    "\n\t movq %2, %%mm1"
		    "\n\t psrad $7, %%mm0"
		    "\n\t psrad $7, %%mm1"
		    "\n\t packssdw %%mm1, %%mm0"
		    "\n\t movq %%mm0, %0"
		    :"=m" (dst[s]) : "m" (xb[0]), "m" (xb[2]));

#define pack16(n) xb[n] >>= 7; clip16(xb[n]); dst[s + n] = xb[n];
#if 0
		{
			int q;
			for (q = 0; q < 4; q++) {
				xb[q] >>= 7;
				clip16(xb[q]);
				assert(dst[s + q] == xb[q]);
			}
		}
		pack16(0);
		pack16(1);
		pack16(2);
		pack16(3);
#endif
	    s += 4;
	}

	for (;s < es; s++) {
	    tmp[0] = 0;
	    for (i = 0; i < info->n_coeff; i++) {
		x = src[s + i * info->channels] * info->coeff[i] >> 8;
		tmp[0] += x;
	    }
	    xb[0] = tmp[0];
	    pack16(0);
	}

	__asm__("\n\t emms\n");
	info->buffer_pos += ticks;
	SINC_ASSERT(info->buffer_pos <= info->buffer_ticks,
		    ("Buffer overrun"));
	return ticks;
}

static int
feed_sinc(struct pcm_feeder	*f,
	  struct pcm_channel	*c,
	  uint8_t		*b,
	  uint32_t		 count,
	  void			*source)
{
	struct feed_sinc_info *info = f->data;
	uint32_t done;
	int fetch, ticks;

	if (info->quality == 0 || info->src == info->dst)
		return FEEDER_FEED(f->source, c, b, count, source);

	done = 0;
	while (done < count) {
		/* Slurp in more data if input buffer is not full */
		fetch = (max_buffer_ticks(info) - info->buffer_ticks) * bytes_per_tick(info);
		if (fetch > 0) {
			uint8_t *u8b;
			u8b = (uint8_t*)info->buffer;
			u8b += (info->buffer_ticks + history_ticks(info)) *
				bytes_per_tick(info);
			fetch = FEEDER_FEED(f->source, c, u8b, fetch, source);
			SINC_ASSERT((fetch % 4) == 0, ("uneven fetch"));
			if (fetch == 0 &&
			    info->buffer_pos == info->buffer_ticks)
				break;
			info->buffer_ticks += fetch / bytes_per_tick(info);
		}

		ticks = (count - done) / bytes_per_tick(info);
		if (ticks > (info->buffer_ticks - info->buffer_pos)) {
			ticks = info->buffer_ticks - info->buffer_pos;
		}
		if (info->impl == 1) {
			done += sinc_filter_mmx(info, ticks, (int16_t*)(b + done))
				* bytes_per_tick(info);
		} else {
			done += sinc_filter(info, ticks, (int16_t*)(b + done))
				* bytes_per_tick(info);
		}

		SINC_ASSERT(done <= count, ("Output buffer overrun"));

		/* Roll history full of data from end of buffer to start */
		if (info->buffer_pos == max_buffer_ticks(info)) {
			uint8_t *u8b;
			u8b = (uint8_t*)info->buffer;
			bcopy(u8b + max_buffer_bytes(info), u8b,
			      history_bytes(info));
			info->buffer_pos = 0;
			info->buffer_ticks = 0;
		}
	}
	return done;
}

static struct pcm_feederdesc feeder_sinc_desc[] = {
	{FEEDER_SINC, AFMT_S16_LE | AFMT_STEREO, AFMT_S16_LE | AFMT_STEREO, 0},
	{0, 0, 0, 0}
};

static kobj_method_t feeder_sinc_methods[] = {
    	KOBJMETHOD(feeder_init,		feed_sinc_init),
    	KOBJMETHOD(feeder_free,		feed_sinc_free),
    	KOBJMETHOD(feeder_set,		feed_sinc_set),
    	KOBJMETHOD(feeder_get,		feed_sinc_get),
    	KOBJMETHOD(feeder_feed,		feed_sinc),
	{ 0, 0 }
};

FEEDER_DECLARE(feeder_sinc, 2, NULL);

