/* -*- c-basic-offset: 8 -*- */
#include <sys/soundcard.h>
#include <sys/ioctl.h>

#include <fcntl.h>
#include <inttypes.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#ifdef __linux__
#include <getopt.h>
#endif

static void
usage(void)
{
        fprintf(stderr, 
		"rawrec [-b bits per sample] [ -c <channels> ] <smpl_rate>");
        exit(EX_USAGE);
}

static int
log_base2(int n)
{
    int u = 0;
    while (n >> u) {
	u++;
    }
    return u;
}

static void
record(const char*	sample_device, 
       uint32_t 	sample_rate, 
       uint32_t 	sample_channels, 
       uint32_t 	sample_encoding,
       int		blksz)
{
	int		fd, frag;
	char		buffer[1024];
	size_t		buffer_bytes = sizeof(buffer) / sizeof(buffer[0]);

	fd = open(sample_device, O_RDONLY);
	if (fd < 0) {
		perror(sample_device);
		exit(EX_OSFILE);
	}
	if (ioctl(fd, SNDCTL_DSP_SETFMT, &sample_encoding) < 0) {
		perror("SNDCTL_DSP_CHANNELS");
		exit(EX_DATAERR);
	}
	if (ioctl(fd, SNDCTL_DSP_CHANNELS, &sample_channels) < 0) {
		perror("SNDCTL_DSP_CHANNELS");
		exit(EX_DATAERR);
	}

	if (ioctl(fd, SNDCTL_DSP_SPEED, &sample_rate) < 0) {
		perror("SNDCTL_DSP_SPEED");
		exit(EX_DATAERR);
	}

	if (blksz > 0) {
		frag = log_base2(blksz) | 0xffff0000;
	
		if (ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &frag) < 0) {
			perror("SNDCTL_DSP_SETFRAGMENT");
			exit(EX_DATAERR);
		}
	}

	fprintf(stderr, "pre-read");
	read(fd, buffer, 4);
	fprintf(stderr, ".\n");

	for(;;) {
		ssize_t rd = read(fd, buffer, buffer_bytes);
		if (rd <= 0) {
			perror("read");
			fprintf(stderr, "read error.");
			break;
		}
		if (fwrite(buffer, 1, rd, stdout) == 0) {
			perror("write");
			break;
		}
	}
}

int main(int argc, char * const argv[]) 
{
	const char* sample_device = "/dev/audio";
	uint32_t sample_rate = 8000;
	uint32_t sample_channels = 1;
	uint32_t sample_encoding = AFMT_S16_LE;
	uint32_t bits = 16;
	int blksz;

	int ch;

	blksz = 512;
	
	while ((ch = getopt(argc, argv, "b:c:f:B:")) != -1) {
		switch(ch) {
		case 'b':	bits = atoi(optarg); break;
		case 'c':	sample_channels = atoi(optarg); break;
		case 'f':	sample_device = optarg; break;
		case 'B':	blksz = atoi(optarg); break;
		}
	}

	argc -= optind;
	argv += optind;

	if (argc != 1)
                usage();

	sample_rate = atoi(argv[0]);
	if (sample_rate < 8000 || sample_rate > 48000) {
		fprintf(stderr, 
			"rate must lie in range 8000-48000\n");
		usage();
	}

	switch (bits) {
	case 8:
		sample_encoding = AFMT_S8;
		break;
	case 16:
		sample_encoding = AFMT_S16_LE;
		break;
	default:
		fprintf(stderr, "bits per sample must be 8 or 16\n");
		usage();
	}

	record(sample_device, sample_rate, sample_channels, sample_encoding,
	       blksz);

	return 0;
}
